import numpy as np
from mag.py.files import linesplit
def from_delim(file:str, delim:str=',', dtype=None, newline='\n') -> list:
    '''Parses deliminated file into numpy array.

    Args:
    '''
    with open(file, 'r') as f:
        data = np.array([linesplit(line, delim, newline) for line in f])

    if dtype is not None:
        data = data.astype(dtype)
    return data



import os
def filelines(filename:str) -> int:
    '''Quickly returns the number of lines in a file.

    Args:
        filename (str): full path to file

    Returns:
        line_no (int): the number of lines if succsesful, None otherwise.
    '''
    with open(filename, 'r') as file:
        return sum(1 for line in file)
    return None



def slices(arr:list, slcs:list) -> list:
    '''Applies a list of slices to a list, joining the results.

    Args:
        arr (list): a list
        slcs (list): a list of slices

    Returns:
        filtered (list): The elements of arr which correspond to slcs
    '''
    return [col for slc in slcs for col in arr[slc]]


def linesplit(line:str, delim:str='\t', newline:str='\n') -> list:
    '''Convention for procesing deliminated lines of a file

    Args:
        line (str): a line of a file
        delim (str): what seperates elements in the line
        newline (str): the newline character

    Returns:
        elements (list): the list of elements of the line split by delim and
            with the newline removed.
    '''
    return line.rstrip(newline).split(delim)

def readsplit(file_object, delim:str='\t', newline:str='\n') -> list:
    '''Convention for procesing deliminated lines of a file

    Args:
        file_object: an opened file from which lines could be read
        delim (str): what seperates elements in the line
        newline (str): the newline character

    Returns:
        elements (list): the list of elements of next line of the file, split by
            delim and with the newline removed.
    '''
    return linesplit(file_object.readline(), delim, newline)


def lineslice(line:str, slcs:list, delim:str='\t', newline:str='\n', add_newline_q:bool=True) -> str:
    '''Convention for extracting elements of a deliminated lines of a file

    Args:
        line (str): a line of a file
        slcs (list): a list of slices of indicating what elements to extract
        delim (str): what seperates elements in the line
        newline (str): the newline character
        add_newline_q (bool): whether or not to add a newline after remerging
            the line.

    Returns:
        filtered_line (str): the string of delim seperated elements from the
            line after being filtered by lists of slcs.
    '''
    return delim.join(slices(linesplit(line, delim, newline), slcs)) \
    + (newline if add_newline_q else '')


def readslice(file_object, slcs:list, delim:str='\t', newline:str='\n', add_newline_q:bool=True) -> str:
    '''Convention for extracting elements of a deliminated lines of a file

    Args:
        file_object: an opened file from which lines could be read
        slcs (list): a list of slices of indicating what elements to extract
        delim (str): what seperates elements in the line
        newline (str): the newline character
        add_newline_q (bool): whether or not to add a newline after remerging
            the line.

    Returns:
        filtered_line (str): the string of delim seperated elements from the
            next line of the file_object after being filtered by lists of
            slcs.
    '''
    return lineslice(file_object.readline(), slcs, delim, newline, add_newline_q)




def multireplace(string:str, replacements:list):
    '''Apply as many replaces to string as specified by replacements

    Args:
        string (str): the string to transform
        replacements (list): a list of (find(str), replace(str)) tuples

    Returns:
        transformed (str): the provided string after sequentially updating
            by calling replace(find, replace) for all elements in replacements

    '''
    transformed = string
    for (find, replace) in replacements:
        transformed.replace(find, replace)
    return transformed


def streamlines(
    file:str,
    keep:list,
    write_dir:str=None,
    prefix:str='',
    suffix:str='',
):
    '''Write the lines specified in <keep> to specified output file.

    Args:
        file (str): full path of a file.
        keep (list): indicies of the lines to keep in the file.
        write_dir (str): directory to write the truncated file. Defaults to same
            directory as file.
        prefix (str): a string to add before the basename of the file in the
            write_dir. Defaults to ''.
        suffix (str): a string to add after the basename of the file in the
            write_dir. Defaults to ''.

    Returns:
        fname (str): the filename where the results were written.

    '''
    if write_dir is None: write_dir = os.path.dirname(file)
    bname, ext = os.path.splitext(os.path.basename(file))
    fname = os.path.join(write_dir, '{}{}{}{}'.format(prefix,bname,suffix,ext))

    with open(file, 'r') as ifile, open(fname, 'w') as ofile:
        for (i, line) in enumerate(ifile):
            if i in keep:
                ofile.write(line)
    return fname

def parse_delim(
    file:str,
    delim:str='\t',
    newline:str='\n',
    types=[]
) -> list:
    '''Convention for procesing deliminated file

    Args:
        file (str): full path to a deliminated file
        delim (str): what seperates elements in the line
        newline (str): the newline character
        types (list): list of tuples (index, type) to cast element_index of each
            line to type
    Returns:
        lines (list): the list of lines whereby elements of the line split by
            the delim with the newline removed.
    '''
    with open(file, 'r') as f:
        lines = []
        for line in f:
            parts = linesplit(line, delim, newline)
            for (i, t) in types:
                parts[i] = t(parts[i])
            lines.append(parts)
        return lines
        # list(map(lambda l: linesplit(l, delim, newline), f.readlines()))

def from_delim(string:str, delim:str='\t', newline:str='\n', types=[])->list:
    '''Convention for procesing deliminated str

    Args:
        string (str): string corresopnding to deliminated file
        delim (str): what seperates elements in the line
        newline (str): the newline character
        types (list): list of tuples (index, type) to cast element_index of each
            line to type
    Returns:
        lines (list): the list of lines whereby elements of the line split by
            the delim with the newline removed.
    '''
    lines = []
    for line in string.rstrip(newline).split(newline):
        parts = linesplit(line, delim, newline)
        for (i, t) in types:
            parts[i] = t(parts[i])
        lines.append(parts)
    return lines


def to_delim(
    iterable:list,
    delim:str='\t',
    newline:str='\n'
) -> str:
    '''Convention for procesing deliminated file

    Args:
        iterable (list): a list of lists corresponding to lines which should be
            joined by delim.
        delim (str): what seperates elements in the line
        newline (str): the newline character

    Returns:
        str: the file as a string
    '''
    return newline.join([delim.join([str(e) for e in sub]) for sub in iterable])



import os, random
def partition_iterable(
    iterable:list,
    train:float=1, valid:float=0, test:float=0,
    shuffle:bool=True
):
    '''Randomly assigns a list of indicies to train, valid and test datasets

    Note: train + valid + test should be 1

    Args:
        iterable (list): an iterable (line numbers of a file, file names, etc).
        train (float): a float in [0, 1] specifying the percent of the files
            should be allocated to the training dataset
        valid (float): a float in [0, 1] specifying the percent of the files
            should be allocated to the training valid
        test (float): a float in [0, 1] specifying the percent of the files
            should be allocated to the training test
        shuffle (boolean): whether or not to shuffle the iterable prior to
            partitioning. Defaults to true.

    Returns:
        datasets (dict): a dictionary with three keys - train, valid, test -
            each of which correpond to a list of elements from iterable.
    '''
    n = len(iterable)

    if shuffle:
        random.shuffle(iterable)

    a = int(n * (train))
    b = int(n * (train + valid))
    c = int(n * (train + valid + test))

    partitioned = {
        'train': set(iterable[:a]),
        'valid': set(iterable[a:b]),
        'test' : set(iterable[b:])
    }
    return partitioned


from mag.py.files import filelines, streamlines
def downsample(
    files:list,
    write_to_disk:bool = False,
    write_dir:str=None,
    prefix:str='',
    suffix:str=None,
) -> list:
    '''Randomly downsamples the provided files so that they contain the same
    number of lines.

    Args:
        files (list): a list of files where it is assumed that each line is a
            record.
        write_to_disk (boolean): whether to write the downsampled lines to disk.
        write_dir (str): where to write the downsampled file. Defaults to same
            directory as the current file being processed in the files list.
        prefix (str): a string to be preprended to each downsampled file.
            Defaults to ''.
        suffix (str): a string to be appened to each downsampled file. Defaults
            to '_downsampled_to_N', where N is the number of lines in the file.

    Returns:
        result (list): if write_to_disk is true, returns the list of output
            filenames, otherwise returns a list of indicies corresponding to the
            randomly downsampled lines for each file.
    '''
    # how many lines per file
    records_per_file = {file: filelines(file) for file in files}
    # minimum lines of all files
    to = min([records for file, records in records_per_file.items()])
    # indicies of all lines per file
    ranges = [list(range(records)) for file, records in records_per_file.items()]
    # mix up lines
    [random.shuffle(rng) for rng in ranges]
    # take minimum lines for each range
    ranges = [set(rng[:to]) for rng in ranges]
    if not write_to_disk:
        return ranges

    filenames = []
    if suffix is None: suffix = '_downsampled_to_{}'.format(to)
    if write_dir is not None: out_dir = write_dir

    for (i, file) in enumerate(files):
        if write_dir is None: out_dir = os.path.dirname(file)
        fname = streamlines(file, ranges[i], out_dir, prefix, suffix)
        filenames.append(fname)

    return filenames




def paritition_class_text_files(
    files:list,
    train:float=1, valid:float=0, test:float=0,
    shuffle:bool=True,
    write_dir:str=None,
)->dict:
    '''Splits each file into train, validation and test sets at the specified
    ratios.

    Args:
        files (list): a list of files where it is assumed that each line is a
            record.
        train (float): a float in [0, 1] specifying the percent of the files
            should be allocated to the training dataset
        valid (float): a float in [0, 1] specifying the percent of the files
            should be allocated to the training valid
        test (float): a float in [0, 1] specifying the percent of the files
            should be allocated to the training test
        shuffle (boolean): whether or not to shuffle the iterable prior to
            partitioning. Defaults to true.
        write_dir (str): the directory where the split files should be written.
            Defaults to the same directory of the original file.

    '''
    filenames = {}
    for file in files:
        n_lines = filelines(file)
        datasets = partition_iterable(list(range(n_lines)), train, valid, test, shuffle)

        bname, ext = os.path.splitext(os.path.basename(file))
        for (ds, keep) in datasets.items():
            fname = streamlines(file, keep, write_dir, '', '_{}'.format(ds))
            if file not in filenames: filenames[file] = {}
            filenames[file][ds] = fname
    return filenames



import tensorflow as tf

def multilabel_loss(inputs, labels):
    with tf.variable_scope('multilabel_loss'):
        return tf.reduce_mean(tf.nn.sigmoid_cross_entropy_with_logits(labels=labels, logits=inputs))



import os, copy, math
import numpy as np

from multiprocessing import Pool
from numpy import mean, subtract
from scipy.spatial.distance import euclidean

from mag.np.arr import runs, binarize, nonzero_1d, consecutive
from .iou import iou_1d
from .alignment import align_channel_objects
from .mask_accuracy import mask_accuracy

'''-----------------------------------------------------------------------------

EXAMPLE'S CHANNEL LEVEL FUNCTIONS

-----------------------------------------------------------------------------'''
def detect_channel_binary_objects(channel:list, cutoff:float=0.5):
    ''' Finds the boundaries of objects in a channel following a binary masking.

    Args:
        channel (list): a list of floats
        cutoff (float): a float in [0, 1] determining which values in channel
            should be transformed to 1 and those which should be 0

    Returns:
        objects (list): a list of objects, [obj1, obj2,...], where each object
            is itself a list [start, stop] indicating the indices in channel
            at which the object starts and stops.

    '''
    # ensure numpy array and apply binary mask
    channel = binarize(np.array(channel), cutoff)
    # get indices of signal
    signal_indicies = nonzero_1d(channel)
    # handle channel with no signal
    if signal_indicies.size == 0: return []
    # get the runs of signal from our channel
    return runs(signal_indicies)




'''-----------------------------------------------------------------------------

EXAMPLE'S LEVEL FUNCTIONS

-----------------------------------------------------------------------------'''
def detect_sequence_binary_objects(channel_matrix, cutoff:float=0.5):
    ''' Finds the boundaries of objects for all channels in a matrix following a binary masking.

    Args:
        channel_matrix (list): a list of list of floats. Each sub-list represents
            a channel.

        cutoff (float): a float in [0, 1] determining which values in channel_matrix
            should be transformed to 1 and those which should be 0

    Returns:
        objects (list): a list of list of objects, [[obj1, obj2,...], [...],...] 
            where each sub-list corresponds to the list of objects for that
            channel where each object is itself a list [start, stop] indicating
            the indices in channel at which the object starts and stops.

    '''
    channel_matrix = binarize(np.array(channel_matrix), cutoff)
    return [detect_channel_binary_objects(channel) for channel in channel_matrix]



def evaluate_segmentation(
    output_channel_matrix,
    target_channel_matrix,
    cutoff:float=0.5,
    alignment_scoring_fn=iou_1d,
    tie_breaking_fn=euclidean,
    alignment_scoring_optimal=min,
    tie_breaking_optimal=min
):
    metrics = {
        'perfect_absences': 0,
        'false_negatives': 0,
        'false_positives': 0,
        'perfect_coverages': 0,
        'aligned': [],
        'offsets': [],
        'ious': [],
        'average_iou': float('nan'),
        'average_offset': [float('nan'), float('nan')]
    }

    additive_metrics = [
        'perfect_absences', 'perfect_coverages',
        'false_negatives', 'false_positives',
        'aligned', 'offsets', 'ious'
    ]

    output_channel_matrix = binarize(np.array(output_channel_matrix), cutoff)
    target_channel_matrix = binarize(np.array(target_channel_matrix), cutoff)

    for channel_index in range(len(target_channel_matrix)):
        # shorter
        channel_str = 'channel_{}_'.format(channel_index)

        # current channel
        output_channel = output_channel_matrix[channel_index]
        target_channel = target_channel_matrix[channel_index]

        # get objects
        output_objects = detect_channel_binary_objects(output_channel, cutoff)
        target_objects = detect_channel_binary_objects(target_channel, cutoff)

        alignment_metrics = align_channel_objects(
            output_objects, target_objects,
            alignment_scoring_fn, tie_breaking_fn,
            alignment_scoring_optimal, tie_breaking_optimal
        )

        for k, v in alignment_metrics.items():
            # store channel wise metrics
            metrics[channel_str+k] = v
            # aggregate metrics of interest
            if k in additive_metrics:
                if type(v) is not list and math.isnan(v): continue
                metrics[k] += v

    if len(metrics['offsets']):
        metrics['average_offset'] = [mean(errors) for errors in zip(*metrics['offsets'])]
    if len(metrics['ious']):
        metrics['average_iou'] = mean(metrics['ious'])

    return metrics


def evaluate_batched_segmentation(
    batched_output_channel_matrix,
    batched_target_channel_matrix,
    cutoff:float=0.5,
    processes:int=None,
    alignment_scoring_fn=iou_1d,
    tie_breaking_fn=euclidean,
    alignment_scoring_optimal=min,
    tie_breaking_optimal=min,
    data_format='channels_first'
):
    metrics = []
    shape = batched_target_channel_matrix.shape
    num_channels = shape[-1] if data_format == 'channels_last' else shape[1]
    if processes is None: processess = os.cpu_count()
    with Pool(processes=processess) as pool:
        starargs = [
            (
                batched_output_channel_matrix[example_index],
                batched_target_channel_matrix[example_index],
                cutoff, alignment_scoring_fn,
                tie_breaking_fn, alignment_scoring_optimal, tie_breaking_optimal
            ) for example_index in range(len(batched_target_channel_matrix))
        ]
        metrics = pool.starmap(evaluate_segmentation, starargs)
    metrics = merge_batched_metrics(metrics, num_channels)
    metrics['accuracy'] = mask_accuracy(
        batched_output_channel_matrix,
        batched_target_channel_matrix,
        cutoff
    )
    return metrics



def merge_batched_metrics(mapped_metrics:dict, num_channels:int):
    merged = {
        'perfect_absences': 0,
        'false_negatives': 0,
        'false_positives': 0,
        'perfect_coverages': 0,
        'aligned': [],
        'offsets': [],
        'ious': [],
        'average_iou': float('nan'),
        'average_offset': [float('nan'), float('nan')]
    }

    _additive_metrics = [
        'perfect_absences', 'perfect_coverages',
        'false_negatives', 'false_positives',
    ]
    _list_metrics = [
        'aligned', 'offsets', 'ious'
    ]

    additive_metrics = copy.deepcopy(_additive_metrics)
    list_metrics = copy.deepcopy(_list_metrics)


    for channel_index in range(num_channels):
        channel_str = 'channel_{}_'.format(channel_index)
        for additive_metric in _additive_metrics:
            additive_metrics.append(channel_str+additive_metric)
            merged[channel_str+additive_metric] = 0
        for list_metric in _list_metrics:
            list_metrics.append(channel_str+list_metric)
            merged[channel_str+list_metric] = []

    for metrics in mapped_metrics:
        for k, v in metrics.items():
            if k in additive_metrics or k in list_metrics:
                if type(v) is not list and math.isnan(v): continue
                merged[k] += v

    if len(merged['offsets']):
        merged['average_offset'] = [mean(errors) for errors in zip(*merged['offsets'])]
    if len(merged['ious']):
        merged['average_iou'] = mean(merged['ious'])

    return merged



'''

index:   0  1  2  3  4  5  6  7  8  9  10 11 12 13 14   [ lower, upper ]
pred1:   |-----------------------|                      [     0,     8 ]
targ1:                     |------------------------|   [     6,    14 ] (expected)

pred2:   |-----------|                                  [     0,     4 ]
targ2:                     |------------------------|   [     6,    14 ] (no overlap)

pred3:   |-----------------------|                      [     0,     8 ]
targ3:                     ||                           [     6,     6 ] (small sequence)

pred4:                     |---------------------|      [     6,    13 ]
targ4:                     |---------------------|      [     6,    13 ] (perfect match)

pred5:                     ||                           [     6,     6 ]
targ5:                     ||                           [     6,     6 ] (perfect match, small sequence)

pred6:                 ||                               [     5,     5 ]
targ6:                     ||                           [     6,     6 ] (mismatch, both small sequence)

pred7:              ||                                  [     4,     4 ]
targ7:                     |------------------------|   [     6,    14 ] (mismatch, one small sequence)


I = min(p_upper, t_upper) - max(p_lower, t_lower)
U = max(p_upper, t_upper) - min(p_lower, t_lower)

i1 = min( 8, 14) - max(0, 6) =  8 - 6 =  2
u1 = max( 8, 14) - min(0, 6) = 14 - 0 = 14  --(IOU)-->  2 / 14 [√] makes sense

i2 = min( 4, 14) - max(0, 6) =  4 - 6 = -2
u2 = max( 4, 14) - min(0, 6) = 14 - 0 = 14  --(IOU)--> -2 / 14 [ ] ERROR should be 0?

i3 = min( 8,  6) - max(0, 6) =  6 - 6 =  0
u3 = max( 8,  6) - min(0, 6) =  8 - 0 =  8  --(IOU)-->  0 /  8 [ ] ERROR, should be 1 / 8

i4 = min(13, 13) - max(6, 6) = 13 - 6 =  7
u4 = max(13, 13) - min(6, 6) = 13 - 6 =  7  --(IOU)-->  7 /  7 [√] makes sense

i5 = min( 6,  6) - max(6, 6) =  6 - 6 =  0
u5 = max( 6,  6) - min(6, 6) =  6 - 6 =  0  --(IOU)-->  0 /  0 [ ] ERROR, just return 0

i6 = min( 5,  6) - max(5, 6) =  5 - 6 = -1
u6 = max( 5,  6) - min(5, 6) =  6 - 5 =  1  --(IOU)--> -1 /  1 [ ] ERROR, should be 0?


i7 = min( 4, 14) - max(4, 6) =  4 - 6 = -2
u7 = max( 4, 14) - min(4, 6) = 14 - 4 = 10  --(IOU)--> -2 / 10 [ ] ERROR should be 0?

We can define 5 unique cases:

1. expected, the sequences overlap (example 1)
2. no overlap (examples 2, 6, 7)
3. perfect (example 4)
4. perfect, small sequences (example 5)
5. expected, but with one small sequence (example 3)

For case 1 and 3 we can trust our formula.
For case 4 we can just return 1
For case 5 we need to add 1 to the intersection
For case 2 we return 0?


'''
def iou_1d(predicted_boundary, target_boundary):
    '''Calculates the intersection over union (IOU) based on a span.

    Notes:
        boundaries are provided in the the form of [start, stop].
        boundaries where start = stop are accepted
        boundaries are assumed to be only in range [0, int < inf)

        if no intersection, returns 0.


    Example:

        0  1  2  3  4  5  6  7  8  9  10 11 12 13 14 15 16 17 18 19 20 index
           |-----------------------|                                   predicted
                             |------------------------|                target

        then the predicted boundaries are:

            [ lower, upper ]
            [     1,     9 ] predicted
            [     7,    15 ] target

        then the intersection of these two sequences would be:

             2 = 9 - 7
               = min(9, 15) - max(1, 7)
               = min(p_upper, t_upper) - max(p_lower, t_lower)

        and the union:

            14 = 15 - 1
               = max(15, 9) - min(7, 1)
               = max(t_upper, p_upper) - min(t_lower, p_lower)

        giving as IOU of:

            0.143 ≈ 1 / 7 =  2 / 14

    Args:
        predicted_boundary (list): the [start, stop] of the predicted boundary
        target_boundary (list): the ground truth [start, stop] for which to compare

    Returns:
        iou (float): the IOU bounded in [0, 1]

    '''

    p_lower, p_upper = predicted_boundary
    t_lower, t_upper = target_boundary

    # boundaries are in form [start, stop] and 0 <= start <= stop
    assert 0<= p_lower <= p_upper
    assert 0<= t_lower <= t_upper

    # no overlap, pred is too far left or pred is too far right
    if p_upper < t_lower or p_lower > t_upper:
        # handle case 2
        return 0

    if predicted_boundary == target_boundary:
        # handle case 3 and 4
        return 1

    intersection_lower_bound = max(p_lower, t_lower)
    intersection_upper_bound = min(p_upper, t_upper)

    intersection = intersection_upper_bound - intersection_lower_bound

    max_upper = max(t_upper, p_upper)
    min_lower = min(t_lower, p_lower)

    union = max_upper - min_lower

    if intersection == 0:
        # handle case 5
        intersection = 1

    # return case 1
    return intersection / union



import tensorflow as tf

def conv1d_chain(inputs, filter_sizes:list, kernel_size:int,  name='conv_1d_chain'):
    with tf.variable_scope(name):
        x = tf.layers.batch_normalization(inputs, reuse=tf.AUTO_REUSE)
        for i, filters in enumerate(filter_sizes):
            x = tf.layers.conv1d(inputs, filters, kernel_size, name='conv_{}'.format(i), padding='same', reuse=tf.AUTO_REUSE)
            x = tf.nn.relu(x)
        return x

def conv1d_to_labels(inputs, labels, kernel_size):
    with tf.variable_scope('conv_to_labels'):
        x = tf.layers.conv1d(inputs, labels, kernel_size, name='conv', padding='same', reuse=tf.AUTO_REUSE)
        return x



import tensorflow as tf
from .branch import branch_a, branch_b

def block_a(inputs, filters, kernel_size, name='block_a'):
    '''1D version of block_a from ademxapp model.

    inputs ---> branch_a ---> outputs
            |              |
            ⌊__ branch_b __⌋
    '''
    with tf.variable_scope(name):
        x = tf.layers.batch_normalization(inputs, name='bn', reuse=tf.AUTO_REUSE)
        x = tf.nn.relu(x)
        b_a = branch_a(x, filters, kernel_size)
        b_b = branch_b(x, filters, kernel_size)
        return (b_a + b_b)

def block_b(inputs, filters, kernel_size, name='block_b'):
    '''1D version of block_b from ademxapp model.

    inputs ---> branch_a ---> outputs
            |              |
            ⌊______________⌋
    '''
    with tf.variable_scope(name):
        x = tf.layers.batch_normalization(inputs, name='bn', reuse=tf.AUTO_REUSE)
        x = tf.nn.relu(x)
        b_a = branch_a(x, filters, kernel_size)
        return (inputs + b_a)

def block(btype, inputs, filters, kernel_size, name='block'):
    blck = block_a if btype.lower() == 'a' else block_b
    name = '{}_{}'.format(name, btype.lower()) if name == 'block' else name
    return blck(inputs, filters, kernel_size, name)



import os

def dir_dirs(directory:str):
    ''' Lists all subdirectorys of the given directory

    Args:
        directory (str): full path to a directory

    Returns:

        subdirs (list): all entries of directory which are also a directory

    '''
    return [
        os.path.join(directory, entry) for entry in os.listdir(directory)
        if os.path.isdir(os.path.join(directory, entry))
    ]



import datetime

def current_date(delim:str='-'):
    '''Returns day, month, and year as a string seperated by delim'''
    now = datetime.datetime.now()
    date = now.strftime('%d'+delim+'%m'+delim+'%Y')
    return date

def current_time(delim:str=':'):
    '''Returns hour and minute as a string seperated by delim'''
    now = datetime.datetime.now()
    time = now.strftime('%H'+delim+'%M')
    return time



import tensorflow as tf
from .block import block

def chain(length, inputs, filters, kernel_size, name='chain', residual_q=True):
    with tf.variable_scope(name):
        x = a = block('a', inputs, filters, kernel_size, 'block_a0')

        for i in range(1, length):
            x = block('b', x, filters, kernel_size, 'block_b{}'.format(i))

        if residual_q:
            x = tf.layers.max_pooling1d(x, pool_size=kernel_size, strides=1, padding='same')

        return (a + x) if residual_q else x


def dist_chain(length, inputs, filters, kernel_size, name='chain', residual_q=True, devices=None):
    if devices is None: return chain(length, inputs, filters, kernel_size, name, residual_q)
    if type(devices) is not list: devices = [devices]

    with tf.variable_scope(name):
        with tf.device(devices[0]):
            x = a = block('a', inputs, filters, kernel_size, 'block_a0')

        for i in range(1, length):
            device = devices[i % len(devices)]
            with tf.device(device):
                x = block('b', x, filters, kernel_size, 'block_b{}'.format(i))

        if residual_q:
            with tf.device(devices[-1]):
                x = tf.layers.max_pooling1d(x, pool_size=kernel_size, strides=1, padding='same')

        return (a + x) if residual_q else x



def tower_chain(length, inputs, filter_sizes, kernel_sizes, name='tower_chain'):
    from .tower import tower
    with tf.variable_scope(name):
        x = inputs
        for fs in filter_sizes:
            x = tower(length, x, fs, kernel_sizes, 'tower_{}'.format(fs))
        return x


def dist_tower_chain(length, inputs, filter_sizes, kernel_sizes, name='tower_chain', devices=None):
    if devices is None: return tower_chain(length, inputs, filter_sizes, kernel_sizes, name)
    from .tower import dist_tower
    with tf.variable_scope(name):
        x = inputs
        for i, filter_size in enumerate(filter_sizes):
            device = devices[i % len(devices)]
            with tf.device(device):
                x = dist_tower(length, x, filter_size, kernel_sizes, 'tower_{}'.format(filter_size), devices)
        return x



def domain(array:list) -> list:
    '''Returns the min and max of the array

    Args:
        array (list): a list of at least one element

    Returns:
        domain (list): [min, max]
    '''
    return [min(array), max(array)]



import tensorflow as tf

def branch_a(inputs, filters, kernel_size,  name='branch_a'):
    with tf.variable_scope(name):
        x = tf.layers.conv1d(inputs, filters, kernel_size, name='conv_1', padding='same', reuse=tf.AUTO_REUSE)
        x = tf.nn.relu(x)
        x = tf.layers.conv1d(x, filters, kernel_size, name='conv_2', padding='same', reuse=tf.AUTO_REUSE)
        return x

def branch_b(inputs, filters, kernel_size, name='branch_b'):
    with tf.variable_scope(name):
        x = tf.layers.conv1d(inputs, filters, kernel_size, name='conv_1', padding='same', reuse=tf.AUTO_REUSE)
        return x



from tensorflow.python.client import device_lib
import os

def available_devices():
    '''

    Returns:
        devices(dict): a list of devices by type (cpu / gpu)
    '''
    devices = device_lib.list_local_devices()
    cpus = [d for d in devices if d.device_type == 'CPU']
    gpus = [d for d in devices if d.device_type == 'GPU']
    return {'cpus': cpus, 'gpus': gpus}


def gpu_available_q():
    '''
    Returns:
        (bool): whether or not there is an availble gpu
    '''
    return len(available_devices()['gpus']) > 0


def set_env_devices(devices:list):
    '''
    Sets the environment variable CUDA_VISIBLE_DEVICES to devices

    Args:
        devices(list): a list of device names

    Returns:
        None

    '''
    os.environ["CUDA_VISIBLE_DEVICES"] = ','.join(devices)



def lookup(dictionary:dict, key, default=None):
    '''Fetches the value of key from dictionary.

    Note:
        this function masks the KeyNotFound error when trying to access the
            a_dict[key] for a key that does not exist, by returning default
            instead.

    Args:
        dictionary (dict): a dict of values

        key: the key for the desired value

        default: what to return if key is not in dictionary. Defaults to None.

    Returns:
        value: either the value of key in dictionary, or default if key is not
            in the dictionary

    '''
    return dictionary[key] if key in dictionary else default

def merge(*dictionaries):
    '''Creates a unified dictionary, overwriting in a leftward fashion
    Args:
        *dictionaries any number of dictionary
    Returns:
        merged (dict): the unified dictionary
    '''
    merged = {}
    for dictionary in dictionaries:
        merged.update(dictionary)
    return merged

def take(keys, dictionary):
    '''Filters dictionary for specified keys

    Args:
        keys (list): a list of dictionary keys
        dictionary (dict): a dictionary of which to filter

    Returns:
        filtered (dict): the provided dictionary filtered to only contain keys.
    '''
    return {k: dictionary[k] for k in filter(lambda k: k in keys, dictionary)}

def drop(keys, dictionary):
    '''Filters dictionary for specified keys

    Args:
        keys (list): a list of dictionary keys
        dictionary (dict): a dictionary of which to filter

    Returns:
        filtered (dict): the provided dictionary filtered to not contain
        keys.
    '''
    return {k: dictionary[k] for k in filter(lambda k: k not in keys, dictionary)}


def safe_merge(*dictionaries):
    '''Created a unified dictionary with at most the keys specified in the first
    dictionary.

    Args:
        *dictionaries any number of dictionary
    Returns:
        merged (dict): the unified dictionary
    '''
    return take(dictionaries[0].keys(), merge(*dictionaries))



import numpy as np

def binarize(data:list, cutoff:float=0.5) -> list:
    '''Converts all values of data to either 1 or 0 depending on
    whether or not the value is greater than cutoff.

    Args:
        data (ndarray): an arbitrary array of numerical values
        cutoff (float): a value bounded [0, 1] determining which values are
            converted to one.

    Returns:
        binary_data (ndarray:int): an array with same dimensions as data where
            all values are either 0 or 1
    '''
    return (data > cutoff).astype(int)

def consecutive(data:list, stepsize:float=1) -> list:
    '''Splits data into subarrays where the values in each array vary by stepsize

    Args:
        data (ndarray): a one dimensional array of sorted values
        stepsize (float): a numeric value specifying how close values should be
            to be considered consecutive
    Returns:
        consecutive_numbers (ndarray): a list of values from data where each
            sublist contains values that are exactly stepsize apart
    '''
    return np.split(data, np.where(np.diff(data) != stepsize)[0]+1)

from mag.py.math import domain
def runs(array:list) -> list:
    '''Returns the [start, stop] values of consectuive numbers in the given array

    Args:
        array (list): a list of values, sorted least to greatest

    Returns:
        runs (list): a list of all runs found in the array, where a run specifies
            the lower and upper bounds of the run
    '''
    return [domain(run) if run.size > 0 else [] for run in consecutive(array)]


def nonzero_1d(data:list) -> list:
    '''Returns the indices of nonzero values in an one-dimensional array.

    Args:
        data (ndarray): an array with one dimension

    Returns:
        indices (ndarray): a one dimensional array containing indicies of data
            which are nonzero
    '''
    return data.nonzero()[0]



import tensorflow as tf
from .chain import chain, dist_chain

def tower(length, inputs, filters, kernel_sizes, name='tower'):
    with tf.variable_scope(name):
        x = tf.layers.batch_normalization(inputs, reuse=tf.AUTO_REUSE)
        chains = [chain(length, x, filters, ks, 'chain_{}'.format(ks)) for ks in kernel_sizes]
        return sum(chains)


def dist_tower(length, inputs, filters, kernel_sizes, name='tower', devices=None):
    if devices is None: return tower(length, inputs, filters, kernel_sizes, name)

    with tf.variable_scope(name):
        x = tf.layers.batch_normalization(inputs, reuse=tf.AUTO_REUSE)
        chains = []
        for i, kernel_size in enumerate(kernel_sizes):
            device = devices[i % len(devices)]
            with tf.device(device):
                cname = 'chain_{}'.format(kernel_size)
                chains.append(dist_chain(length, x, filters, kernel_size, cname, devices=[device]))
        return sum(chains)



from itertools import combinations

def subsets(setnames:list) -> list:
    ''' Gives the 2^n subsets for every set in setnames

    Args:
        setnames (list): a list of elements, where each element corresponds to
            the name of a set. e.g. ['A', 'B', ..., 'Z']
    Returns:
        subsets (list): a list of tuples, where each tuple contains the set
            names corresponding to the sets in that subset

            e.g. [(), (A, ), (B, ), ..., (A, B, ), ... (A, B, ..., Z)]
    '''
    number_of_sets = len(setnames)
    return [
        combination
        for combination_index in range(number_of_sets)
        for combination in list(combinations(setnames, combination_index))
    ] + [tuple(setnames)]

def partitions_by_length(partitions:list):
    '''Groups partitions by number of sets in each partition

    Args:
        partitions (list): a list of lists, where each sublist represents a
            partititon and each item therein corresponds to the name of a set.

            e.g. the output of subsets

    Returns:

        partitons_by_length (dict): {num: partitons} pairs where num is an int
            specifiy the number of sets in each partition, and partitions is
            a list containing all partitions with num sets.

    '''
    grouped = {}
    for partition in partitions:
        cardinality = len(partition)
        if cardinality not in grouped:
            grouped[cardinality] = []
        grouped[cardinality].append(partition)
    return grouped

def upset_elements(sets:dict, domain:set=set({})) -> dict:
    '''Gives the elements unique to each ~region~ of the venn diagram

    Args:
        sets (dict): the pair of {setname (str): elements (list)} for which to
            compute on

        domain (set): all the elements, including those which may not be seen in
            the passed sets. By default, an empty set.

    Returns:
        elements_unique_to_comparisons (dict): a dict containing the pair of
            {comparison (tuple): elements (list)}, where comparison is a tuple
            of setnames (keys from sets), e.g. ('A', 'B',) indicating the
            comparison made.

    '''
    setnames = list(sets.keys())
    # returns partitions of setnames in increasing number of sets, e.g.
    # [(), (A, ), (B, ), ..., (A, B, ), ... (A, B, ..., Z)]
    comparisons = subsets(setnames)
    comparisons.reverse()

    # group each comparison by how many sets in each comparisons e.g.
    # work our way from the inner most of the venn diagram to the outer most
    grouped_comparisons = partitions_by_length(comparisons)

    # results
    elements_unique_to_comparisons = {}

    # elements we have already seen
    seen = set({})

    # start with the intersection of all sets and work our way to the empty set
    for n_sets in range(len(setnames), -1, -1):

        partitions_with_n_sets = grouped_comparisons[n_sets]
        current_comparisons = [
            [
                # list of elements for each set in the current comparison
                sets[name] for name in comparison
            ]  for comparison in partitions_with_n_sets
            # where each partition of n sets is a comparison to make
        ]

        # the empty set
        if partitions_with_n_sets == [()]:
            # elements unique to the empty set is anything in the domain not
            # already seen
            elements_unique_to_comparisons.update({
                (): domain.difference(seen)
            })
            # nothing else to do
            continue

        intersections = [
            # elements shared across the sets in the comparison
            set.intersection(*comparison_sets)
            # for all comparisons with n sets in them
            for comparison_sets in current_comparisons
        ]

        uniques = [
            # filter out elements already seen
            intersection.difference(seen)
            for intersection in intersections
        ]

        elements_unique_to_comparisons.update(
            dict(
                list(
                    zip(
                        partitions_with_n_sets, uniques
                    ) # tuples of (sets in current partition, unique elements)
                ) # cast as a list
            ) # cast as a dict
        ) # update our results

        just_seen = set.union(*uniques)
        seen.update(just_seen)

    return elements_unique_to_comparisons



import json
def load_to_local(file:str) -> dict:
    '''
    Reads the passed file and dumps all keys into locals()

    Args:
        file (str): fullpath to a simple key: value json file

    Returns:
        data (dict): the contents of the json file
    '''
    with open(file, 'r') as f:
        data = json.load(f)
        for var, val in data.items():
            locals()[var] = val
        return data



from mag.np.arr import runs, binarize, nonzero_1d, consecutive
import numpy as np

def mask_accuracy(outputs, targets, cutoff:float=0.5):
    '''Calculates the one-to-one accuracy of the binary mask with the specified
    cuttoff of the predictions to the labels

    Args:
        predictions (np.ndarray): an arbitrary array, with the same shape as
            labels containing the predicted probabilities of each position
            belonging the corresponding label.

        labels (np.ndarray): an arbitrary array of binary values

        cutoff (float): a value bounded by [0, 1] specifying the minimum
            probability required to consider a position belong to the label of
            a given class.

    Returns:
        accuracy (float): how many binarized predictions corresponding directly
            to labels
    '''
    b_mask = binarize(outputs, cutoff)
    return  (b_mask == targets).sum() / len(targets.flatten())



import numpy as np
from numpy import mean, subtract
from scipy.spatial.distance import euclidean
from .iou import iou_1d

def align_channel_object(
    output_object:list,
    target_objects:list,
    alignment_scoring_fn=iou_1d,
    tie_breaking_fn=euclidean,
    alignment_scoring_optimal=min,
    tie_breaking_optimal=min
) -> int:
    '''Attepts to align a given object to a list of other objects

    Args:
        output_object (list): an object in form [start, stop]

        target_objects (list): a list of objects in form [[start, stop],...]

        alignment_scoring_fn (function): a function which is passed the
            output_object and a target_object and should return a singular
            numeric value. Defaults to iou_1d

        tie_breaking_fn (function): a function which is passed the output_object
            and a target_object and should return a singular numeric value.
            Defaults to euclidean. Only called if there are more than one optimal
            alignment score.

        alignment_scoring_optimal (function): a function which returns the optimal
            alignment score from a list of scores. Defaults to min.

        tie_breaking_optimal (function): a function which returns the optimal
            tie breaking score from a list of scores. Defaults to min.


    Returns:
        alignment_index (int): the index of the object in target_objects which
            is the best alignment for output_object given the alignment_scoring
            and  putatively invoked tie_breaking functions

    '''
    # calculate output object's alignment scores for every target object
    alignment_scores = [
        alignment_scoring_fn(output_object, target_object)
        for target_object in target_objects
    ]

    # find the best score
    best_score = alignment_scoring_optimal(alignment_scores)

    # find occurances of the best
    occurances = np.where(np.array(alignment_scores) == best_score)[0]

    # only one optimal aligment
    if occurances.size == 1:
        alignment_index = occurances[0]

    # tie breaking can drastically improve alignment results
    else:
        # calculate the tie breaking scores only for the best scoring targets
        tie_breaking_scores = [
            tie_breaking_fn(output_object, target_object)
            for i, target_object in enumerate(target_objects)
            if i in occurances
        ]
        # find the new best score
        best_score = tie_breaking_optimal(tie_breaking_scores)

        alignment_index = tie_breaking_scores.index(best_score)
        alignment_index = occurances[alignment_index]

    # return the alignment index
    return alignment_index


def align_channel_objects(
    output_objects, target_objects,
    alignment_scoring_fn=iou_1d,
    tie_breaking_fn=euclidean,
    alignment_scoring_optimal=min,
    tie_breaking_optimal=min
) -> dict:
    '''Attepts to align each object of a list of objects to a list of different objects

    Args:
        output_objects (list): a list of objects in form [[start, stop],...]

        target_objects (list): a list of objects in form [[start, stop],...]

        alignment_scoring_fn (function): a function which is passed the
            output_object and a target_object and should return a singular
            numeric value. Defaults to iou_1d

        tie_breaking_fn (function): a function which is passed the output_object
            and a target_object and should return a singular numeric value.
            Defaults to euclidean. Only called if there are more than one optimal
            alignment score.

        alignment_scoring_optimal (function): a function which returns the optimal
            alignment score from a list of scores. Defaults to min.

        tie_breaking_optimal (function): a function which returns the optimal
            tie breaking score from a list of scores. Defaults to min.


    Returns:
        metrics (dict): the metrics of alignment. Current metrics include:

            perfect_absences, perfect_coverages, false_negatives, false_positives,
            aligned, offsets, ious, average_iou, average_offset

    '''
    metrics = {
        'perfect_absences': 0,
        'false_negatives': 0,
        'false_positives': 0,
        'perfect_coverages': 0,
        'aligned': [],
        'offsets': [],
        'ious': [],
        'average_iou': float('nan'),
        'average_offset': [float('nan'), float('nan')]
    }

    number_of_output_objects = len(output_objects)
    number_of_target_objects = len(target_objects)

    # at least one object in both output and target
    is_alignable       = all([number_of_output_objects, number_of_target_objects])
    # no objects in the output or the target
    is_perfect_absence = (number_of_output_objects == number_of_target_objects == 0)
    # no objects in the output or but objects in the target
    is_false_negative  = (not is_perfect_absence and number_of_output_objects == 0)
    # no objects in the target or but objects in the output
    is_false_positive  = (not is_perfect_absence and number_of_target_objects == 0)

    if is_perfect_absence: metrics['perfect_absences'] += 1
    if is_false_negative:  metrics['false_negatives']  += number_of_target_objects
    if is_false_positive:  metrics['false_positives']  += number_of_output_objects

    if is_alignable:
        for output_object in output_objects:
            aligned_index  = align_channel_object(
                output_object, target_objects,
                alignment_scoring_fn, tie_breaking_fn,
                alignment_scoring_optimal, tie_breaking_optimal
            )

            aligned_object = target_objects[aligned_index]

            metrics['aligned'] += [(output_object, aligned_object)]

            if output_object == aligned_object: metrics['perfect_coverages'] += 1

            offset = subtract(output_object, aligned_object)
            iou = iou_1d(output_object, aligned_object)

            metrics['offsets'].append(offset)
            metrics['ious'].append(iou)

    if len(metrics['offsets']):
        metrics['average_offset'] = [mean(errors) for errors in zip(*metrics['offsets'])]
    if len(metrics['ious']):
        metrics['average_iou'] = mean(metrics['ious'])
    return metrics
